#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/keysym.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "x-props.h"
#include "txtren.h"
#include "graymap.h"
#include "list.h"

#include "config.h"
#include "medit.h"
#include "appmenu.h"
#include "imgload.h"
#include "x-keyboard.h"
#include "txent.h"
#include "btn.h"
#include "chfile.h"
#include "chicon.h"
#include "chdir.h"
#include "label.h"
#include "scan_appdir.h"

typedef struct xmedit menu_editor_t;


enum { 
  BTN_TLEFT1,  BTN_TRIGHT1, 
  BTN_TLEFT2,  BTN_TRIGHT2, 
  BTN_ROOT1,   BTN_UP1,       BTN_TRASH1,
  BTN_ROOT2,   BTN_UP2,       BTN_TRASH2,
  BTN_DELETE1, BTN_EDIT1,     BTN_ARROW1,
  BTN_DELETE2, BTN_EDIT2,     BTN_ARROW2,
  BTN_FOLDER1, BTN_ITEM1,     BTN_SEP1,
  BTN_FOLDER2, BTN_ITEM2,     BTN_SEP2,
  BTN_DONE,    BTN_SCAN,
  BTN_MOVEUP1, BTN_MOVEDOWN1, BTN_SORT1, BTN_OPEN1,
  BTN_MOVEUP2, BTN_MOVEDOWN2, BTN_SORT2, BTN_OPEN2,
};

typedef struct xmpane
{
  struct xmpane *other;
  menu_editor_t *editor;
  list_t *list;
  list_item_t *selected;
  amenu_t *current;
} xmpane_t;

struct xmedit
{
  Window      window;
  ximage_t   *image;
  int         clicked[5];
  amenu_t    *root, *hidden;
  amenu_t    *tbleft, *tbright;
  xmpane_t    left, right;
  cfg_top_t  *cfg;
  char       *fname;
  void      (*func)(void*);
  void       *arg;
};

typedef struct { unsigned int iw, ih, *idata; int blticon; } item_aux_t;


typedef struct 
{
  Window parent; int parent_w, parent_h;
  Window window;
  ximage_t *image;
  aitem_t *item;
  text_entry_t *te[3];
  int focus;
  x_keyboard_t *kb;
  int clicked[5];
  menu_editor_t *editor;
} item_editor_t;

static item_editor_t *item_editor_new(Window parent, 
                               int parent_w, int parent_h,
                               aitem_t *item, menu_editor_t *me);


static int btn_pos[]= {
   4,   5,  62,  24, BTN_TLEFT1,
  69,   5,  62,  24, BTN_TRIGHT1,
 509,   5,  62,  24, BTN_TLEFT2,
 574,   5,  62,  24, BTN_TRIGHT2,

   4,  32,  62,  24, BTN_ROOT1,
  69,  32,  62,  24, BTN_UP1,
 134,  32,  62,  24, BTN_TRASH1,
 444,  32,  62,  24, BTN_ROOT2,
 509,  32,  62,  24, BTN_UP2,
 574,  32,  62,  24, BTN_TRASH2,

   4, 396, 62, 24, BTN_SORT1,
 134, 396, 62, 24, BTN_MOVEUP1,
 199, 396, 62, 24, BTN_MOVEDOWN1,

 574, 396, 62, 24, BTN_SORT2,
 379, 396, 62, 24, BTN_MOVEUP2,
 444, 396, 62, 24, BTN_MOVEDOWN2,

   4, 424, 62, 24, BTN_DELETE1,
  69, 424, 62, 24, BTN_OPEN1,
 134, 424, 62, 24, BTN_EDIT1,
 199, 424, 62, 24, BTN_ARROW1,
 379, 424, 62, 24, BTN_ARROW2,
 444, 424, 62, 24, BTN_EDIT2,
 509, 424, 62, 24, BTN_OPEN2,
 574, 424, 62, 24, BTN_DELETE2,

 293, 436, 62, 24, BTN_DONE,
 293, 408, 62, 24, BTN_SCAN,

  69, 452, 62, 24, BTN_ITEM1,
 134, 452, 62, 24, BTN_FOLDER1,
 199, 452, 62, 24, BTN_SEP1,
 379, 452, 62, 24, BTN_SEP2,
 444, 452, 62, 24, BTN_FOLDER2,
 509, 452, 62, 24, BTN_ITEM2 };


#include "ame-scr1.c"

static void me_expose(menu_editor_t *me, XExposeEvent *ev);
static void me_press(menu_editor_t *me, XButtonEvent *ev);
static void me_release(menu_editor_t *me, XButtonEvent *ev);
static void me_client(menu_editor_t *me, XClientMessageEvent *ev);
static void me_destroy(menu_editor_t *me, XEvent *ev);

static void display_folder(xmpane_t *pane, amenu_t *menu);
static amenu_t *find_parent_menu(amenu_t *hay, amenu_t *needle);
static void move_item(xmpane_t *src, xmpane_t *dst);
static void init_pane(menu_editor_t *me, xmpane_t *self, xmpane_t *other,
                      int x,int y, int w, int h);
static void delete_item(xmpane_t *src,xmpane_t *dst);

static void move_to_hidden(amenu_t *M, xmpane_t *dst);
static void destroy_aitem(aitem_t *it);
static void move_item_up(xmpane_t *pane);
static void move_item_down(xmpane_t *pane);
static void new_item(xmpane_t *pane,aitem_t *it);
static void sort_list(xmpane_t *pane);
static list_item_t *mk_listitem(aitem_t *it);

static void edit_item(xmpane_t *pane, list_item_t *li);
static void pane_update_item(xmpane_t *pane, aitem_t *ai);

static void scan_screen(menu_editor_t *me);
static char* path_to_folder(menu_editor_t *me,amenu_t *m);
static void amenu_free_all(amenu_t *M);

static void add_config_program(amenu_t *A)
{
  aitem_t *T;
  T= amenu_item_new(A, "MuPanel Settings", ":self-settings.png",0);
  T->type= 0;
  T->exec= strdup(":configure");
  T->src= strdup(":internal");
}



void edit_app_menu
    (Window parent,int x,int y,char *fname, cfg_top_t *cfg,
     void (*func)(void*), void *arg)
{
  menu_editor_t *me;
  int w,h;
  XSetWindowAttributes attribs;
  attribs.background_pixmap= None;

  w= bi_medit_width; h= bi_medit_height;

  me= malloc(sizeof(*me));
  me->window= XCreateWindow(x_display, parent, x,y, w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  XSelectInput(x_display, me->window, ExposureMask | 
                       ButtonPressMask | ButtonReleaseMask);

  register_event_handler(me->window, Expose, me, me_expose, NULL);
  register_event_handler(me->window, ButtonPress, me, me_press, NULL);
  register_event_handler(me->window, ButtonRelease, me, me_release, NULL);
  register_event_handler(me->window, ClientMessage, me, me_client, NULL);

  set_window_property_atoms(me->window, a_wm_protocols,
                               1,&a_wm_delete_window);
  set_window_property_string(me->window, a_wm_name, "MuPanel Settings");

  me->image= create_image(w, h);
  blend_alpha_image(me->image->data, w, h, 0,0,
      (uint32_t*)   bi_medit, w,h, 0,0, w,h, 0x0);

  me->root= read_app_menu(fname, &me->hidden);
  me->tbleft= read_app_menu(cfg->left_box_file,NULL);
  me->tbright= read_app_menu(cfg->right_box_file,NULL);
  me->cfg= cfg;
  me->fname= strdup(fname);
  me->func= func;
  me->arg= arg;

  if (me->root->items.next==&me->root->items)
     add_config_program(me->root);    

  init_pane(me, &me->left, &me->right,   5,  88, 309, 303);
  init_pane(me, &me->right, &me->left, 327,  88, 308, 303);

  XMapWindow(x_display, me->window);
}

static void me_expose(menu_editor_t *me, XExposeEvent *ev)
{
  put_image(me->window, ev->x, ev->y,
            me->image, ev->x, ev->y,
            ev->width, ev->height);
}

static void me_press(menu_editor_t *me, XButtonEvent *ev)
{
  btn_click(me->window, me->image, btn_pos, sizeof(btn_pos)/sizeof(int), 
                        ev, me->clicked);
}

static void me_release(menu_editor_t *me, XButtonEvent *ev)
{
  int k;
  xmpane_t *pane;
  amenu_t *parent;
  aitem_t *it;
  void (*F)(void*); void *A;
  Window window;

  k= btn_release(me->window, me->image, ev, me->clicked);
  switch(k)
  {
  case -1: return ; // nothing clicked
  case BTN_TRASH1: case BTN_ROOT1: case BTN_UP1:
  case BTN_ARROW1: case BTN_DELETE1:
  case BTN_MOVEUP1: case BTN_MOVEDOWN1: 
  case BTN_SEP1: case BTN_OPEN1: case BTN_SORT1:
  case BTN_FOLDER1:  case BTN_ITEM1:
  case BTN_EDIT1: case BTN_TLEFT1: case BTN_TRIGHT1:
    pane= &me->left; break;
  case BTN_ROOT2: case BTN_TRASH2: case BTN_UP2:
  case BTN_ARROW2: case BTN_DELETE2:
  case BTN_MOVEUP2: case BTN_MOVEDOWN2:  
  case BTN_SEP2: case BTN_OPEN2: case BTN_SORT2:
  case BTN_FOLDER2: case BTN_ITEM2:
  case BTN_EDIT2:
  case BTN_TLEFT2: case BTN_TRIGHT2:
    pane= &me->right; break;
  case BTN_SCAN:
    scan_screen(me);
    return ;
  case BTN_DONE:
    if (me->fname)
    {
      write_app_menu(me->fname, me->root, me->hidden);
      free(me->fname);
    }
    write_app_menu(me->cfg->left_box_file, me->tbleft, NULL);
    write_app_menu(me->cfg->right_box_file, me->tbright, NULL);
    F= me->func;
    A= me->arg;
    window= me->window;
    me_destroy(me, NULL);
    XDestroyWindow(x_display, window);
    if (F) (*F)(A);
    return ; 
  default: printf("clicked button %d\n", k); break;
  }

  switch(k)
  {
  case BTN_FOLDER1: case BTN_SEP1:
  case BTN_FOLDER2: case BTN_SEP2:
  case BTN_UP1: case BTN_UP2:
    if (pane->current==me->hidden || pane->current==me->tbleft
        || pane->current==me->tbright)
            return ;
  case BTN_ITEM1: case BTN_ITEM2:
    if (pane->current==me->hidden)
        return ;
  }

  switch(k)
  {
  case BTN_TLEFT1: case BTN_TLEFT2: display_folder(pane, me->tbleft); break;
  case BTN_TRIGHT1: case BTN_TRIGHT2: display_folder(pane, me->tbright); break;
  case BTN_TRASH1: case BTN_TRASH2: display_folder(pane, me->hidden); break;
  case BTN_ROOT1: case BTN_ROOT2: display_folder(pane, me->root);  break;
  case BTN_UP1: case BTN_UP2:
    if (pane->current==me->root) return ;
    parent= find_parent_menu(pane->editor->root,pane->current);
    if (parent) display_folder(pane, parent);
    break;
  case BTN_ARROW1: case BTN_ARROW2: move_item(pane, pane->other); break;
  case BTN_DELETE1: case BTN_DELETE2: delete_item(pane,pane->other); break;
  case BTN_MOVEUP1: case BTN_MOVEUP2: move_item_up(pane); break;
  case BTN_MOVEDOWN1: case BTN_MOVEDOWN2: move_item_down(pane); break;
  case BTN_SEP1: case BTN_SEP2:
    it= calloc(1,sizeof(*it)); it->type= 2;
    new_item(pane,it);
    break;
  case BTN_FOLDER1: case BTN_FOLDER2:
    it= calloc(1,sizeof(*it));
    it->type= 1;
    it->name= strdup("New Folder");
    it->icon= strdup(":folder.png");
    it->sub= malloc(sizeof(*it->sub));
    it->sub->items.prev= it->sub->items.next= &it->sub->items;
    new_item(pane,it);
    break;
  case BTN_ITEM1: case BTN_ITEM2:
    it= calloc(1,sizeof(*it));
    it->type= 0;
    it->name= strdup("New Command");
    it->icon= strdup(":default-button.png");
    new_item(pane,it);
    break;
  case BTN_OPEN1: case BTN_OPEN2:
    if (!pane->selected) return;
    it= list_item_data(pane->selected);
    if (it->type!=1) return ;
    if (it->sub) display_folder(pane, it->sub);
    break;
  case BTN_SORT1: case BTN_SORT2: sort_list(pane); break;
  case BTN_EDIT1: case BTN_EDIT2:
    if (!pane->selected) return ;
    edit_item(pane, pane->selected);
    break;
  }
}

static void me_destroy(menu_editor_t *me, XEvent *ev)
{
  destroy_image(me->image);
  list_free(me->left.list);
  list_free(me->right.list);
  unregister_event_handler(me->window, Expose, NULL);
  unregister_event_handler(me->window, ButtonPress, NULL);
  unregister_event_handler(me->window, ButtonRelease, NULL);
  amenu_free_all(me->root);
  amenu_free_all(me->hidden);
  amenu_free_all(me->tbleft);
  amenu_free_all(me->tbright);
  free(me);
}

static void me_client(menu_editor_t *me, XClientMessageEvent *ev)
{
  if (ev->message_type==a_wm_protocols && ev->data.l[0]==a_wm_delete_window)
  {
    me_destroy(me, NULL);
    XDestroyWindow(x_display, ev->window);
    return ;
  }
}

static void scan_complete(menu_editor_t *me)
{
  display_folder(&me->left, me->root);
  display_folder(&me->right, me->root);
}

static amenu_t *find_parent_menu(amenu_t *hay, amenu_t *needle)
{
  aitem_t *it;
  amenu_t *R;
  for(it=hay->items.next;it!=&hay->items;it=it->next)
  {
    if (it->type!=1) continue;
    if (it->sub==needle) return hay;
    R= find_parent_menu(it->sub, needle);
    if (R) return R;
  }
  return NULL;
}

static void make_item_aux(aitem_t *it)
{
  item_aux_t *au;
  au= calloc(1,sizeof(*au));
  if (it->icon && it->icon[0])
   load_icon(it->icon, 16, 16, &au->iw, &au->ih, &au->idata);
  it->aux= au;
  if (it->icon && it->icon[0]==':') au->blticon= 1;
}

static void display_folder(xmpane_t *pane, amenu_t *menu)
{
  aitem_t *it;
  list_t *list;
  menu_editor_t *me;

  list= pane->list;
  if (!menu) return ;
  list_clear(list);
 
  for(it=menu->items.next;it!=&menu->items;it=it->next)
    list_append(list, mk_listitem(it));
  pane->selected= NULL;
  pane->current= menu;

  int P;
  static const int labpos[]= { 5,  60, 234,  22, 401,  60, 234,  22 };
  char *path;

  me= pane->editor;
  if (pane==&me->left) P= 0; else P= 4;

  path= path_to_folder(me, menu);
  display_string_lxr(me->window, me->image, labpos+P, path, 0, 0xffffff);
  free(path);
}

static void click_item(xmpane_t *pane, list_t *list, list_item_t *item)
{
  if (pane->selected==item) return ;
  if (pane->selected) list_item_select(list, pane->selected,0); 
  pane->selected= item;
  list_item_select(list, item, 1);
}

static void dlick_item(xmpane_t *pane, list_t *list, list_item_t *item)
{
  aitem_t *it;
  it= list_item_data(item);
  switch(it->type)
  {
  case 0: edit_item(pane, item); break;
  case 1: display_folder(pane, it->sub); break;
  case 2: break; // printf("double click separator\n"); break;
  }
}

static void edit_item(xmpane_t *pane, list_item_t *li)
{
  aitem_t *it;
  it= list_item_data(li);
  if (it->type==2) return ;
  item_editor_new(pane->editor->window, 
                  pane->editor->image->width,
                  pane->editor->image->height, it, pane->editor);
}

static void init_pane(menu_editor_t *me, xmpane_t *self, xmpane_t *other,
                      int x,int y, int w, int h)
{
  self->list= list_new(me->window, x,y,w,h, 
                             click_item, dlick_item, self);
  self->other= other;
  self->selected= NULL;
  self->editor= me;
  display_folder(self, me->root);
}

static int is_parent(amenu_t *hay, amenu_t *A)
{
  aitem_t *it;
  for(it=hay->items.next;it!=&hay->items;it=it->next)
  {
    if (it->type!=1) continue;
    if (it->sub==A) return 1;
    if (is_parent(it->sub, A)) return 1;
  }
  return 0;
}

static void move_item(xmpane_t *src, xmpane_t *dst)
{
  list_item_t *li;
  aitem_t *it;
  amenu_t *da;

  li= src->selected;
  if (!li || src->current==dst->current) return ;
  it= list_item_data(li);
  if (it->type==2) return ;
  if (it->type==1 && it->sub==dst->current) return ;
  if (it->type==1 && is_parent(it->sub,dst->current)) return ;
  list_remove(src->list, li);
  list_append(dst->list, li);

  it->prev->next= it->next;
  it->next->prev= it->prev;

  da= dst->current;

  it->next= &da->items;
  it->prev= da->items.prev;
  da->items.prev->next= it;
  da->items.prev= it;
  src->selected= NULL;
}

static void delete_item(xmpane_t *src,xmpane_t *dst)
{
  menu_editor_t *me;
  list_item_t *li,*di;
  aitem_t *it;

  me= src->editor;
  if (src->current==me->hidden) return ;
  if (!src->selected) return ;

  li= src->selected;
  it= list_item_data(li);

  if (it->type==2) 
  {
    if (dst->current==src->current)
    {
      di= list_find_item(dst->list,it);
      if (dst->selected==di) dst->selected= NULL;
      list_remove(dst->list, di);
      list_item_free(di);
    }
    goto done;
  }

  if (dst->current!=me->hidden) display_folder(dst, me->hidden);
  if (it->type==0) { move_item(src, dst); return ; }

  move_to_hidden(it->sub, dst);

done:
  list_remove(src->list, li);
  src->selected= NULL;
  list_item_free(li);
  it->prev->next= it->next;
  it->next->prev= it->prev;
  destroy_aitem(it);
}

static void move_to_hidden(amenu_t *M, xmpane_t *dst)
{
  aitem_t *it,*N;
  amenu_t *D;

  D= dst->current;

  for(it=M->items.next;it!=&M->items;it=N)
  {
    N= it->next;
    switch(it->type)
    {
    case 1:
      move_to_hidden(it->sub, dst); 
    case 2:
      destroy_aitem(it);
      break;
    case 0: 
      list_append(dst->list, mk_listitem(it));

      it->prev= D->items.prev;
      it->next= &D->items;
      D->items.prev->next= it;
      D->items.prev= it;
      break;
    }
  }
  M->items.prev= M->items.next= &M->items;
  free(M);
}

static void destroy_item_aux(aitem_t *it)
{
  item_aux_t *aux;
  aux= it->aux;
  if (aux)
  {
    if (aux->idata && !aux->blticon) free(aux->idata);
    free(aux);
  }
  it->aux= NULL;
}

static void destroy_aitem(aitem_t *it)
{
  destroy_item_aux(it);
  if (it->name) free(it->name);
  if (it->icon) free(it->icon);
  if (it->exec) free(it->exec);
  if (it->src) free(it->src);
  free(it);
}

static void move_item_up(xmpane_t *pane)
{
  aitem_t *it,*A;
  list_item_t *an, *li;

  li= pane->selected;
  if (!pane->selected) return ;
  
  it= list_item_data(li);
  A= it->prev;
  if (A==&pane->current->items) return ;

  an= list_find_item(pane->list, A);
  if (!an) return ;
  list_remove(pane->list, li);
  list_insert_before(pane->list, an, li);
  list_item_select(pane->list, li, 1);

  if (pane->current==pane->other->current)
  {
    pane= pane->other;
    li= list_find_item(pane->list,it);
    if (!li) goto done;
    an= list_find_item(pane->list, A);
    if (!an) goto done;
    list_remove(pane->list, li);
    list_insert_before(pane->list, an, li);
    if (li==pane->selected)
      list_item_select(pane->list, li, 1);
  }
done:
  it->prev= A->prev;
  A->next= it->next;

  A->next->prev= A;
  it->prev->next= it;

  A->prev= it;
  it->next= A;
}

static void move_item_down(xmpane_t *pane)
{
  aitem_t *it,*A;
  list_item_t *an, *li;

  li= pane->selected;
  if (!pane->selected) return ;
  
  it= list_item_data(li);
  A= it->next;
  if (A==&pane->current->items) return ;

  an= list_find_item(pane->list, A);
  if (!an) return ;
  list_remove(pane->list, li);
  list_insert_after(pane->list, an, li);
  list_item_select(pane->list, li, 1);

  if (pane->current==pane->other->current)
  {
    pane= pane->other;
    li= list_find_item(pane->list,it);
    if (!li) goto done;
    an= list_find_item(pane->list, A);
    if (!an) goto done;
    list_remove(pane->list, li);
    list_insert_after(pane->list, an, li);
    if (li==pane->selected)
      list_item_select(pane->list, li, 1);
  }
done:
  it->next= A->next;
  A->prev= it->prev;

  A->prev->next= A;
  it->next->prev= it;

  A->next= it;
  it->prev= A;
}

static void new_item(xmpane_t *pane, aitem_t *it)
{
  aitem_t *an;
  list_item_t *ali,*bli;
  amenu_t *M;

  M= pane->current;

  if (!pane->selected) { an= &M->items; goto append_stuff; }

  bli= NULL;
  ali= pane->selected;
  an= list_item_data(ali);
  if (pane->other->current==pane->current)
  {
    bli= list_find_item(pane->other->list, an);
    if (!bli) goto append_stuff;
  }
  list_insert_before(pane->list, ali, mk_listitem(it));
  if (bli)
    list_insert_before(pane->other->list, bli, mk_listitem(it));
  goto insert_node;
  
append_stuff:
  list_append(pane->list,mk_listitem(it));
  if (pane->other->current==pane->current)
    list_append(pane->other->list, mk_listitem(it));
insert_node:
  it->prev= an->prev;
  it->next= an;
  an->prev->next= it;
  an->prev= it;
}

static int exchange_items(aitem_t *A, aitem_t *B)
{
  aitem_t TA, TB;

  TA= *A;
  TB= *B;

  *A= TB; *B= TA;

  A->prev= TA.prev; A->next= TA.next;
  B->prev= TB.prev; B->next= TB.next;
  return 1;
}

static int compare_items(aitem_t *A, aitem_t *B)
{
  static const int item_prec[3]= { 1, 0, 2 };
  if (A->type!=B->type) return item_prec[A->type] - item_prec[B->type];
  if (A->name) { if (B->name) return strcmp(A->name, B->name); 
                         else return 1; }
  if (B->name) return -1;
  return 0;
}

static void sort_list(xmpane_t *pane)
{
  aitem_t *it;
  int change;
  aitem_t *head;

  head= &pane->current->items;
  do{
    change= 0;
    for(it=head->next;it->next!=head;it=it->next)
      if (compare_items(it,it->next) > 0)
         change+= exchange_items(it, it->next);
  } while(change);

  display_folder(pane, pane->current);
  if (pane->current==pane->other->current)
     display_folder(pane->other, pane->other->current);
}

static list_item_t *mk_listitem(aitem_t *it)
{
  item_aux_t *aux;
  if (it->type==2) return list_item_new(NULL, 0,0, "---Separator---",it);
  if (!it->aux) make_item_aux(it);
  aux= it->aux;
  return list_item_new(aux->idata, aux->iw, aux->ih, it->name, it);
}

static void pane_update_item(xmpane_t *pane, aitem_t *it)
{
  list_item_t *li,*ni;
  li= list_find_item(pane->list, it);
  if (!li) return ;

  ni= mk_listitem(it);
  list_insert_before(pane->list, li, ni);
  list_remove(pane->list, li);
  list_item_free(li);
  if (pane->selected==li)
  {
    pane->selected= ni;
    list_item_select(pane->list, ni, 1);
  } 
}

static int search_for_name(amenu_t *src, amenu_t *tgt, char **R, int pos)
{
  aitem_t *it;
  int len,Lnm;
  char *nm;

  if (!src || !tgt) return 0;
  for(it=src->items.next;it!=&src->items;it=it->next)
  {
     if (it->type!=1) continue;
     if (it->name) nm= it->name;
              else nm= "?";
     Lnm= strlen(nm);
     if (it->sub==tgt)
     {
       len= pos + 1 + Lnm + 1;
       *R= malloc(len);
       (*R)[len-1]= 0;
       goto write_it;
     }
     if (search_for_name(it->sub, tgt, R,pos+strlen(nm)+1))
        goto write_it;
  }
  return 0;

write_it:
  (*R)[pos]= '/';
  memcpy(*R + pos + 1, nm, Lnm);
  return 1;
}

static char* path_to_folder(menu_editor_t *me,amenu_t *m)
{
  char *R; 
  if (m==me->hidden)  return strdup("Trash");
  if (m==me->root)    return strdup("/");
  if (m==me->tbleft)  return strdup("Taskbar Left");
  if (m==me->tbright) return strdup("Taskbar Right");
  R= NULL;
  if (search_for_name(me->root, m, &R, 0)) return R;
  return strdup("??");
}

static void freenonull(void *k) { if (k) free(k); }

static void amenu_free_all(amenu_t *M)
{
  aitem_t *it, *N;
  if (!M) return ;
  for(it=M->items.next;it!=&M->items;it=N)
  {
     N= it->next;
     if (it->type==1) amenu_free_all(it->sub);
     destroy_item_aux(it);
     freenonull(it->name);
     freenonull(it->icon);
     freenonull(it->exec);
     freenonull(it->src);
     free(it);
  }
  free(M);
}

/***********************************************************************
 *                                                                     *
 *                             Item Editor                             *
 *                                                                     *
 ***********************************************************************/
#include "ame-scr2.c"

static void ie_expose(item_editor_t *ie, XExposeEvent *ev);
static void ie_press(item_editor_t *ie, XButtonEvent *ev);
static void ie_release(item_editor_t *ie, XButtonEvent *ev);
static void ie_key(item_editor_t *ie, XEvent *ev);
static void ie_destroy(item_editor_t *ie, XEvent *ev);

static const int ie_butpos[]=
{
   238, 259, 74, 26, 'O',
   321, 259, 74, 26, 'C',
   468, 191, 74, 26, 'I',
   468, 221, 74, 26, 'E',
};

static const int ie_tpos[]=
{
  162,162,300,24,
  162,192,300,24,
  162,222,300,24,
};

static item_editor_t *item_editor_new(Window parent, 
                               int parent_w, int parent_h,
                               aitem_t *item,menu_editor_t *me)
{
  item_editor_t *ie;
  int w,h;
  XSetWindowAttributes attribs;
  attribs.background_pixmap= None;

  w= bi_iedit_width; h= bi_iedit_height;

  ie= calloc(1,sizeof(*ie));
  ie->parent= parent;
  ie->parent_w= parent_w;
  ie->parent_h= parent_h;
  ie->editor= me;

  ie->window= XCreateWindow(x_display, parent, (parent_w-w)/2,
                                               (parent_h-h)/2, w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  ie->image= create_image(w, h);
  XSelectInput(x_display, ie->window, ExposureMask |  StructureNotifyMask |
                       ButtonPressMask | ButtonReleaseMask | KeyPressMask);

  register_event_handler(ie->window, Expose, ie, ie_expose, NULL);
  register_event_handler(ie->window, ButtonPress, ie, ie_press, NULL);
  register_event_handler(ie->window, ButtonRelease, ie, ie_release, NULL);
  register_event_handler(ie->window, KeyPress, ie, ie_key, NULL);
  register_event_handler(ie->window, DestroyNotify, ie, ie_destroy, NULL);

  blend_alpha_image(ie->image->data, w, h, 0,0,
      (uint32_t*)   bi_iedit, w,h, 0,0, w,h, 0x0);

  ie->item= item;
  ie->te[0]= text_entry_init(ie->window, ie->image,
                      ie_tpos[0],ie_tpos[1],ie_tpos[2],ie_tpos[3],
                               item->name, 1);
  ie->te[1]= text_entry_init(ie->window, ie->image,
                      ie_tpos[4],ie_tpos[5],ie_tpos[6],ie_tpos[7],
                               item->icon, 0);
  if (item->type==0)
  {
    ie->te[2]= text_entry_init(ie->window, ie->image,
                      ie_tpos[8],ie_tpos[9],ie_tpos[10],ie_tpos[11],
                               item->exec, 0);
  }
  else
  {
    fill_rect(ie->image->data, w,h, 73, 221, 469, 26, 0xffffff);
  }
  ie->focus= 0;

  ie->kb= x_keyboard_init(ie->window);

  XMapWindow(x_display, ie->window);
  return ie;
}


static void ie_expose(item_editor_t *ie, XExposeEvent *ev)
{
  put_image(ie->window, ev->x, ev->y, ie->image, ev->x, ev->y,
            ev->width, ev->height);
}
static void ie_press(item_editor_t *ie, XButtonEvent *ev)
{
  int np,i,f;
  if (ie->item->type==1) np= 15; else np= 20;
  if (btn_click(ie->window, ie->image, ie_butpos, np, ev, ie->clicked)) 
     return ;
  np= 4*np/5;
  for(i=0;i<np;i+=4)
    if (ev->x>=ie_tpos[i+0] && ev->x<ie_tpos[i]+ie_tpos[i+2] &&
        ev->y>=ie_tpos[i+1] && ev->y<ie_tpos[i+1] +ie_tpos[i+3])
    {
      f= i/4;
      if (ie->focus==f) break; 
      text_entry_focus(ie->te[ie->focus], 0);
      ie->focus= f;
      text_entry_focus(ie->te[ie->focus], 1);
      break;
    }
}

static void change_exe(void *arg, char *str)
{
  item_editor_t *ie;
  ie= arg;
  if (ie->te[2]) text_entry_set(ie->te[2], str);
}

static void change_icon(void *arg, char *str)
{
  item_editor_t *ie;
  ie= arg;
  if (ie->te[1]) text_entry_set(ie->te[1], str);
}

static void ie_destroy(item_editor_t *ie, XEvent *ev)
{
  int i;
  x_keyboard_free(ie->kb);
  for(i=0;i<3;i++) if (ie->te[i]) text_entry_free(ie->te[i]);
  destroy_image(ie->image);
  unregister_event_handler(ie->window, Expose, NULL);
  unregister_event_handler(ie->window, ButtonPress, NULL);
  unregister_event_handler(ie->window, ButtonRelease, NULL);
  unregister_event_handler(ie->window, KeyPress, NULL);
  unregister_event_handler(ie->window, DestroyNotify, NULL);
  free(ie);
}

static void ie_free(item_editor_t *ie)
{
  Window window;
  window= ie->window;
  ie_destroy(ie,NULL);
  XDestroyWindow(x_display,window);
}

static void repl_str(char **dst, char *src)
{
  if (!src) return ;
  if (!src[0]) { free(src); return ; }
  if (*dst) free(*dst);
  *dst= src;
}

static void ie_done(item_editor_t *ie)
{
  aitem_t *it;
  menu_editor_t *me;

  it= ie->item;

  if (ie->te[0]) repl_str(&it->name, text_entry_value(ie->te[0]));
  if (ie->te[1]) repl_str(&it->icon, text_entry_value(ie->te[1]));
  if (it->type==0 && ie->te[2])
     repl_str(&it->exec, text_entry_value(ie->te[2]));

  me= ie->editor;
  destroy_item_aux(it);
  make_item_aux(it);
  pane_update_item(&me->left, it);
  pane_update_item(&me->right, it);
  ie_free(ie);
}

static void ie_release(item_editor_t *ie, XButtonEvent *ev)
{
  switch(btn_release(ie->window, ie->image, ev, ie->clicked))
  {
  case 'I':
    choose_icon(ie->window, 50, 50, 550, 420, "/usr/share/icons",
                change_icon, ie);
    break;
  case 'E':
    choose_file(ie->window, 50, 50, 550, 420, NULL, "/usr/bin",
                change_exe, ie);
    break;
  case 'C': ie_free(ie); break;
  case 'O': ie_done(ie); break;
  }
}
static void ie_key(item_editor_t *ie, XEvent *ev)
{
  int nte;
  uint32_t *str;size_t nstr; KeySym sym;

  if (!x_keyboard_event(ie->kb, ev, &sym, &str, &nstr)) return ;

  switch(sym)
  {
  case XK_Tab:
        if (ie->item->type==0) nte=3; else nte=2;
        text_entry_focus(ie->te[ie->focus], 0);
        ie->focus= (ie->focus+1)%nte;
        text_entry_focus(ie->te[ie->focus], 1);
        break;
  case XK_Return:
        ie_done(ie);
        break; 
  case XK_Escape:
        ie_free(ie);
        break;
  default:
        text_entry_key(ie->te[ie->focus], sym, str, nstr);
        break;
  }
}

/***********************************************************************
 *                                                                     *
 *                             Scan Window                             *
 *                                                                     *
 ***********************************************************************/


static const int ds_butpos[]= {
 212, 396, 67, 52, 'S',
 285, 396, 67, 24, 'U',
 356, 396, 67, 24, 'D',
 427, 396, 67, 24, 'R',
 498, 396, 67, 24, 'A',
 569, 424, 67, 24, 'B',
   5, 452, 67, 24, 'Q' };

static const int ds_widgetpos[]= {
   5,  72, 630, 317,   // List
   4, 396,  79,  24,   // Shortcut
  87, 396,  79,  24,   // Icon
 286, 425, 278,  22,   // Entry
 286, 454, 349,  22 }; // Scanning Dir

typedef struct 
{
  Window         window;
  ximage_t      *image;
  menu_editor_t *editor;
  int            w,h;
  x_keyboard_t  *kb;
  list_t        *list;
  int            clicked[5];
  text_entry_t  *entry;
  int            active;
  int            n_dirs[2];
  int            s_dirs[2];
  char         **dirs[2];
  list_item_t   *selected;
} dirscan_t;

static void ds_expose(dirscan_t *ds, XExposeEvent *ev);
static void ds_press(dirscan_t *ds, XButtonEvent *ev);
static void ds_release(dirscan_t *ds, XButtonEvent *ev);
static void ds_destroy(dirscan_t *ds, XEvent *ev);
static void ds_key(dirscan_t *ds, XEvent *ev);

static void fill_default_scandirs(cfg_top_t *cfg);
static void ds_display_list(dirscan_t *ds);
static void ds_click_item(dirscan_t *ds, list_t *list,list_item_t *it);
static void ds_chosen_dir(void *arg, char *dir);

#include "ame-scr3.c"

static void scan_screen(menu_editor_t *me)
{
  dirscan_t *ds;
  int w,h;
  XSetWindowAttributes attribs;

  attribs.background_pixmap= None;

  w= bi_scan_width; h= bi_scan_height;

  ds= calloc(1,sizeof(*ds));
  ds->window= XCreateWindow(x_display, me->window, 0,0, w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  XSelectInput(x_display, ds->window, ExposureMask |  KeyReleaseMask |
                           StructureNotifyMask |
                       ButtonPressMask | ButtonReleaseMask | KeyPressMask);

  register_event_handler(ds->window, Expose, ds, ds_expose, NULL);
  register_event_handler(ds->window, ButtonPress, ds, ds_press, NULL);
  register_event_handler(ds->window, ButtonRelease, ds, ds_release, NULL);
  register_event_handler(ds->window, KeyPress, ds, ds_key, NULL);
  register_event_handler(ds->window, KeyRelease, ds, ds_key, NULL);
  register_event_handler(ds->window, DestroyNotify, ds, ds_destroy, NULL);

  ds->w= w;
  ds->h= h;
  ds->image= create_image(w, h);
  ds->kb= x_keyboard_init(ds->window);

  ds->entry= text_entry_init(ds->window, ds->image, 
             ds_widgetpos[12], ds_widgetpos[13],
             ds_widgetpos[14], ds_widgetpos[15],
             "", 1);

  fill_default_scandirs(me->cfg);
  ds->n_dirs[0]= me->cfg->N_menu_shortcut_path;
  ds->s_dirs[0]= me->cfg->S_menu_shortcut_path;
  ds->dirs[0]= me->cfg->menu_shortcut_path;

  ds->n_dirs[1]= me->cfg->N_menu_icon_path;
  ds->s_dirs[1]= me->cfg->S_menu_icon_path;
  ds->dirs[1]= me->cfg->menu_icon_path;

  ds->active= 0;
  ds->editor= me;

  ds->list= list_new(ds->window,
       ds_widgetpos[0], ds_widgetpos[1], 
       ds_widgetpos[2], ds_widgetpos[3], 
       ds_click_item, NULL, ds);

  blend_alpha_image(ds->image->data, w, h, 0,0,
      (uint32_t*)   bi_scan, w,h, 0,0, w,h, 0xffffff);

  ds_display_list(ds);

  XMapWindow(x_display, ds->window);
}


static void ds_expose(dirscan_t *ds, XExposeEvent *ev)
{
  put_image(ds->window, ev->x, ev->y, ds->image, ev->x, ev->y,
            ev->width, ev->height);
}

static int in_rect_ptr(XButtonEvent *ev, const int *pos)
{
  if (ev->x<pos[0]) return 0;
  if (ev->y<pos[1]) return 0;
  if (ev->x>=pos[0]+pos[2]) return 0;
  if (ev->y>=pos[1]+pos[3]) return 0;
  return 1;
}

static void ds_press(dirscan_t *ds, XButtonEvent *ev)
{
  int ac;
  int x,y,w,h;
  if (btn_click(ds->window, ds->image, ds_butpos, 35, ev, ds->clicked)) 
     return ;
  ac= -1;
  if (in_rect_ptr(ev, ds_widgetpos+4)) ac= 0;
  else if (in_rect_ptr(ev, ds_widgetpos+8)) ac= 1;

  if (ac==-1) return ;
  
  if (ds->active==ac) return ;
  x= ds_widgetpos[4]; y= ds_widgetpos[5];
  w= ds_widgetpos[8]+ds_widgetpos[10] - x;
  h= ds_widgetpos[9]+ds_widgetpos[11] - y;
  xor_rect(ds->image->data, ds->w, ds->h, 
               ds_widgetpos[4], ds_widgetpos[5], 
               ds_widgetpos[6], ds_widgetpos[7], 0xffffff);
  xor_rect(ds->image->data, ds->w, ds->h, 
               ds_widgetpos[8], ds_widgetpos[9], 
               ds_widgetpos[10], ds_widgetpos[11], 0xffffff);
  XClearArea(x_display, ds->window, x,y,w,h, True);
  ds->active= ac;
  ds_display_list(ds);
}

static int find_in_array(dirscan_t *ds, char *k)
{
  int i;
  for(i=0;i<ds->n_dirs[ds->active];i++)
    if (ds->dirs[ds->active][i]==k)
       return i;
  return -1;
}

static void ds_release(dirscan_t *ds, XButtonEvent *ev)
{
  int k,i,j,a;
  char *d, *p,*txt;
  list_item_t *pi;
  char *dn;
  cfg_top_t *cfg;
  menu_editor_t *me;
  Window window;

  k= btn_release(ds->window, ds->image, ev, ds->clicked);
  switch(k)
  {
  default: printf("button %c\n",k);
  case -1: return ;
  case 'U':
    if (!ds->selected) return ;
    d= list_item_data(ds->selected);
    i= find_in_array(ds, d);
    if (i<=0) return ;
    p= ds->dirs[ds->active][i-1];
    pi= list_find_item(ds->list, p);
    if (!pi) return ;
    list_remove(ds->list, ds->selected);
    list_insert_before(ds->list, pi, ds->selected);
    list_item_select(ds->list, ds->selected,1);

    ds->dirs[ds->active][i-1]= d;
    ds->dirs[ds->active][i]= p;
    break;
  case 'D':
    if (!ds->selected) return ;
    d= list_item_data(ds->selected);
    i= find_in_array(ds, d);
    if (i<0 || i==ds->n_dirs[ds->active]-1) return ;
    p= ds->dirs[ds->active][i+1];
    pi= list_find_item(ds->list, p);
    if (!pi) return ;
    list_remove(ds->list, ds->selected);
    list_insert_after(ds->list, pi, ds->selected);
    list_item_select(ds->list, ds->selected,1);

    ds->dirs[ds->active][i+1]= d;
    ds->dirs[ds->active][i]= p;
    break;
  case 'R':
    if (!ds->selected) return ;
    d= list_item_data(ds->selected);
    i= find_in_array(ds, d);
    if (i<0) return ;
    list_remove(ds->list, ds->selected);
    ds->selected= NULL;

    free(d);
    for(j=i;j<ds->n_dirs[ds->active]-1;j++)
      ds->dirs[ds->active][j]= ds->dirs[ds->active][j+1];
    ds->n_dirs[ds->active]--;
    break;
  case 'A':
    a= ds->active;
    txt= text_entry_value(ds->entry);
    if (txt[0]==0) { free(txt); return ; }
    if (ds->n_dirs[a]==ds->s_dirs[a])
    {
      ds->s_dirs[a] += ds->s_dirs[a]/2;
      ds->dirs[a]= realloc(ds->dirs[a], sizeof(char*)*ds->s_dirs[a]);
    }
    ds->dirs[a][ds->n_dirs[a]++]= txt;
    list_append(ds->list, list_item_new(NULL,0,0, txt, txt));
    break;
  case 'B':
    if (ds->active) dn= "/usr/share/icons";
              else  dn= "/usr/share/applications";
    choose_dir(ds->window, dn, ds_chosen_dir, ds);
    break;
  case 'S':
    display_string_lxr(ds->window, ds->image, ds_widgetpos+16, 
                        "Scanning", 0x0, 0xffffff);
    put_image(ds->window, ds_widgetpos[16], ds_widgetpos[17], 
              ds->image,  ds_widgetpos[16], ds_widgetpos[17], 
                          ds_widgetpos[18], ds_widgetpos[19]);
    XSync(x_display, False);
    scan_app_dirs(ds->dirs[0], ds->n_dirs[0],
                  ds->dirs[1], ds->n_dirs[1],
                  ds->editor->root, ds->editor->hidden);
    display_string_lxr(ds->window, ds->image, ds_widgetpos+16, 
                        "Scan complete", 0x0, 0xffffff);
    break;
  case 'Q':
    cfg= ds->editor->cfg;
    cfg->N_menu_shortcut_path= ds->n_dirs[0];
    cfg->S_menu_shortcut_path= ds->s_dirs[0];
    cfg->menu_shortcut_path= ds->dirs[0];

    cfg->N_menu_icon_path= ds->n_dirs[1];
    cfg->S_menu_icon_path= ds->s_dirs[1];
    cfg->menu_icon_path= ds->dirs[1];

    window= ds->window;
    me= ds->editor;

    list_free(ds->list);
    ds_destroy(ds, NULL);
    XDestroyWindow(x_display, window);
    scan_complete(me);
    break;
  }
}

static void ds_destroy(dirscan_t *ds, XEvent *ev)
{
  unregister_event_handler(ds->window, Expose, NULL);
  unregister_event_handler(ds->window, ButtonPress, NULL);
  unregister_event_handler(ds->window, ButtonRelease, NULL);
  unregister_event_handler(ds->window, KeyPress, NULL);
  unregister_event_handler(ds->window, KeyRelease, NULL);
  unregister_event_handler(ds->window, DestroyNotify, NULL);

  destroy_image(ds->image);
  text_entry_free(ds->entry);
  x_keyboard_free(ds->kb);
  free(ds);
  // printf("scanner destroyed\n");
}

static void ds_key(dirscan_t *ds, XEvent *ev)
{
  uint32_t *str;size_t nstr; KeySym sym;

  if (!x_keyboard_event(ds->kb, ev, &sym, &str, &nstr)) return ;

  switch(sym)
  {
  case XK_Tab:
  case XK_Return:
  case XK_Escape:
        break;
  default:
        text_entry_key(ds->entry, sym, str, nstr);
        break;
  }
}

static void mk_strarray(char ***Rarray, int *Rn, int *Rs, 
                        int n_initial, char **initial)
{
  char **array;
  int n,s,i;

  array= *Rarray;
  n= *Rn;
  s= *Rs;

  if (array)
  {
   for(i=0;i<n;i++) if (array[i]) free(array[i]);
   free(array);
  }

  s= n_initial + 10;
  array= malloc(sizeof(char*)*s);
  for(i=0;i<n_initial;i++)
    array[i]= strdup(initial[i]);

  *Rarray= array;
  *Rn= n_initial;
  *Rs= s;
}

static void fill_default_scandirs(cfg_top_t *cfg)
{
  static char* icons[]= {
             "/usr/share/icons/hicolor",
             "/usr/share/icons/gnome"};
  static char* apps[]= { "/usr/share/applications" };
  if (cfg->N_menu_icon_path==0)
    mk_strarray(&cfg->menu_icon_path, 
           &cfg->N_menu_icon_path, &cfg->S_menu_icon_path,
            2, icons);
  if (cfg->N_menu_shortcut_path==0)
    mk_strarray(&cfg->menu_shortcut_path, 
           &cfg->N_menu_shortcut_path, &cfg->S_menu_shortcut_path,
             1, apps);
}

static void ds_display_list(dirscan_t *ds)
{
  int a;
  int i;
  a= ds->active;
  list_clear(ds->list);
  ds->selected= NULL;
  for(i=0;i<ds->n_dirs[a];i++)
    list_append(ds->list,
       list_item_new(NULL, 0,0, ds->dirs[a][i], ds->dirs[a][i]));
}

static void ds_click_item(dirscan_t *ds, list_t *list,list_item_t *li)
{
  if (ds->selected==li) return ;
  if (ds->selected)
   list_item_select(list, ds->selected,0);
  list_item_select(list, li, 1);
  ds->selected= li;
}

static void ds_chosen_dir(void *arg, char *dir)
{
  dirscan_t *ds;
  ds= arg;
  if (!dir) return ;
  text_entry_set(ds->entry, dir);
}
