#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-image.h"
#include "txtren.h"
#include "config.h"
#include "clock.h"
#include "graymap.h"
#include "x-events.h"

static Window    cl_window;
static ximage_t *cl_image;
static u_font_t *cl_font;
static int       cl_w, cl_h;
static uint32_t  CL_bg          = 0xededed;
static uint32_t  CL_fg          = 0;
static uint32_t  CL_font_size   = 12;
static char     *CL_font_name   = ":default";
static int       CL_pad         = 3;

static time_t   last_time;  // in minutes

static int get_latest_time();
static void render_clock();
static void clock_expose(void *, XExposeEvent *);

void init_clock(Window parent, int parent_h, int parent_y, 
                cfg_clock_t *cfg, Window *Rwindow, int *Rw)
{
  uint32_t tw; uint8_t *td;

  if (cfg->G_bg)        CL_bg= cfg->bg;
  if (cfg->G_fg)        CL_fg= cfg->fg;
  if (cfg->G_font_size) CL_font_size= cfg->font_size;
  if (cfg->G_font_name) CL_font_name= cfg->font_name;
  if (cfg->G_pad)       CL_pad= cfg->pad;

  cl_font= font_new(CL_font_name, CL_font_size);
  text_render(cl_font, "99:99", &tw, &td);
  free(td);

  cl_w= tw + 2*CL_pad;
  cl_h= parent_h;
  cl_image= create_image(cl_w, cl_h);
  cl_window= create_window_Ob(parent, 0,0, cl_w, cl_h, ExposureMask, 0);

  register_event_handler(cl_window, Expose, NULL, clock_expose, NULL);
  *Rwindow= cl_window;
  *Rw= cl_w;

  get_latest_time();
  render_clock();
}

static int get_latest_time()
{
  struct timeval tv;
  int R;
  gettimeofday(&tv, NULL);
  R= (tv.tv_sec/60!=last_time/60);
  last_time= tv.tv_sec;
  return R;
}

static void render_clock()
{
  uint8_t *td;
  uint32_t tw,th;
  int x,y, W,H;
  char buf[200];
  struct tm* tm;
  fill_rect(cl_image->data, cl_w, cl_h, 0, 0,cl_w, cl_h, CL_bg);

  tm= localtime(&last_time);
  
  sprintf(buf,"%d:%02d", tm->tm_hour, tm->tm_min);
  text_render(cl_font, buf, &tw, &td);

  th= cl_font->height;

  if (cl_w>tw) x= (cl_w - tw) / 2; else x= 0;
  if (cl_h>th) y= (cl_h - th) / 2; else y= 0;

  W= tw;
  H= th;

  graymap_draw(cl_image->data, cl_w, cl_h, x,y,
                   td, tw, th, 0,0,
                CL_fg, CL_bg, W,H);
  free(td);
  XClearArea(x_display, cl_window, 0,0, cl_w, cl_h, True);
}

void update_clock()
{
  if (get_latest_time()) render_clock();
}


static void clock_expose(void *ignore, XExposeEvent *ev)
{
  put_image(cl_window, ev->x, ev->y,
            cl_image, ev->x, ev->y,
            ev->width, ev->height);
}
