#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-misc-props.h"
#include "x-image.h"
#include "x-events.h"
#include "config.h"
#include "clients.h"
#include "taskbar.h"
#include "graymap.h"
#include "panel.h"

/****************************************************************
****************************************************************/

static ximage_t *pager_image;
static Window pager_window;
static uint32_t n_desktops, current_desktop;
static int *pager_w_tgt;

static int root_height;

/****************************************************************
  The colors below are in rgb format. 
****************************************************************/
static uint32_t pc_border          = 0xededed;
static uint32_t pc_a_root          = 0x376594;
static uint32_t pc_a_window        = 0x85bcf6;
static uint32_t pc_a_window_border = 0xffffff;
static uint32_t pc_i_root          = 0xa6a6a6;
static uint32_t pc_i_window        = 0xededed;
static uint32_t pc_i_window_border = 0x313739;

/****************************************************************
  Pager dimensions. 
****************************************************************/
static int page_w, page_h;
static int pager_w, pager_h;


static void calculate_pager_dimensions(int maxh);
static void draw_client_icon
               (client_window_t *cw,int x,int y,int w,int h,uint32_t bg);
static void render_pager();

static void pager_click(void *, XButtonEvent *);
static void pager_expose(void *, XExposeEvent *);
static void root_property_change(void *ignore, XPropertyEvent *pe);

static void pgcolor(uint32_t *C, int G, uint32_t V)
{
  if (G) *C= V;
}

static void init_pager_colors(cfg_pager_style_t *cfg)
{
  pgcolor(&pc_border, cfg->G_border, cfg->border);
  pgcolor(&pc_a_root, cfg->active.G_root, cfg->active.root);
  pgcolor(&pc_a_window, cfg->active.G_window, cfg->active.window);
  pgcolor(&pc_a_window_border, cfg->active.G_window_border,
                               cfg->active.window_border);
  pgcolor(&pc_i_root, cfg->inactive.G_root, cfg->inactive.root);
  pgcolor(&pc_i_window, cfg->inactive.G_window, cfg->inactive.window);
  pgcolor(&pc_i_window_border, cfg->inactive.G_window_border,
                               cfg->inactive.window_border);
}

void init_pager
  (Window parent, int parent_h, int parent_y, cfg_pager_style_t *cfg,
   Window *Rwindow, int *Rw)
{
  root_height= parent_y;
  init_pager_colors(cfg);
  if (get_number_of_desktops(&n_desktops)) n_desktops= 1;
  if (get_current_desktop(&current_desktop)) current_desktop= 1;
  if (n_desktops>20) n_desktops= 20;

  calculate_pager_dimensions(parent_h);

  pager_image= create_image(pager_w, pager_h);
  pager_window= create_window_Ob(parent, 0,0, pager_w, pager_h,
                                 ButtonPressMask | ExposureMask, 0);
  *Rw= pager_w; *Rwindow= pager_window;

  render_pager();

  register_event_handler(pager_window, Expose, NULL, pager_expose, NULL);
  register_event_handler(pager_window, ButtonPress, NULL, pager_click, NULL);
  register_event_handler(x_root_window, PropertyNotify, NULL,
                                root_property_change, NULL);

  XSelectInput(x_display, x_root_window, PropertyChangeMask );
                                         //ButtonPressMask  );

  pager_w_tgt= Rw;
}

static void calculate_pager_dimensions(int maxh)
{
  page_h= maxh - 2;
  page_w= (double) x_root_width / root_height * page_h;
  pager_h= maxh;
  pager_w= page_w * n_desktops + (n_desktops+1);
}


static void render_pager_background()
{
  int i;
  draw_rect( pager_image->data, pager_w, pager_h,
                  0, 0, pager_w, pager_h, pc_border);
  for(i=1;i<n_desktops;i++)
    draw_vline(pager_image->data, pager_w, pager_h,
                 i*(page_w+1), 1, pager_h-1, pc_border);

  for(i=0;i<n_desktops;i++)
     fill_rect(pager_image->data, pager_w, pager_h,
                    1+i*(page_w+1), 1, page_w, page_h, 
                    i==current_desktop ? pc_a_root : pc_i_root);
}

static void render_mini_window(Window window)
{
   int w_x, w_y, w_w, w_h;
   int desk_s, desk_e, desk;
   int p_x, p_y, p_w, p_h;
   client_window_t *cw;

   cw= get_client(window);
   if (!cw) return ;

   if (get_client_state(cw) || cw->state) 
     { //printf("skipping window %d because of state\n", (int) cw->id); 
       return ; }
   if (!get_client_type(cw) && cw->type) 
     { //printf("skipping window %d because of type\n", (int) cw->id); 
       return ; }
   if (get_client_pos(cw))
        { w_x= 0; w_y= 0; w_w= x_root_width; w_h= root_height; }
   else { w_x= cw->x; w_y= cw->y; w_w= cw->width; w_h= cw->height; }

   if (get_client_desktop_number(cw) || cw->desk<0 || cw->desk>=n_desktops)
        { desk_s= 0; desk_e= n_desktops-1; }
   else desk_s= desk_e= cw->desk;

   get_client_icon(cw, 16,16); // FIXME: not 25x25, something flexible.

   if (w_x<0) { w_w+= w_x; w_x= 0; }
   if (w_y<0) { w_h+= w_y; w_y= 0; }

   if (w_x+w_w>x_root_width) w_w= x_root_width - w_x;
   if (w_y+w_h>root_height) w_h= root_height - w_y;

   //printf("window %d w_x= %d w_y= %d w_w= %d w_h= %d\n",
   //         (int) cw->id, w_x, w_y, w_w, w_h);

   p_x= (double) w_x / x_root_width * page_w;
   p_y= (double) w_y / root_height * page_h;
   p_w= (double) w_w / x_root_width * page_w;
   p_h= (double) w_h / root_height * page_h;

   if (p_w<3 || p_h<3) return ;

   for(desk=desk_s;desk<=desk_e;desk++)
   {
      uint32_t C_wb, C_wi;
      int D_px, D_py;

      D_px= 1 + desk*(page_w+1) + p_x;
      D_py= 1 + p_y;
      C_wb= (desk==current_desktop ?  pc_a_window_border: pc_i_window_border);
      C_wi= (desk==current_desktop ?  pc_a_window: pc_i_window);

      draw_rect(pager_image->data, pager_w, pager_h,
                     D_px, D_py, p_w, p_h, C_wb);

      fill_rect(pager_image->data, pager_w, pager_h,
                     D_px+1, D_py+1, p_w-2, p_h-2, C_wi);

      draw_client_icon( cw, D_px, D_py, p_w, p_h, C_wi);
   }
}

static void render_pager()
{
  int i;
  Window *windows;
  size_t n_windows;

  render_pager_background();
  if (get_window_property_windows(x_root_window, 
                                  a_net_client_list_stacking,
                                   &n_windows, &windows, 0))
     return ;
  for(i=0;i<n_windows;i++)
     render_mini_window(windows[i]);
  free(windows);
}


static void draw_client_icon
     (client_window_t *cw,int x,int y,int w,int h,uint32_t bg)
{
  if ((cw->flags&CWF_ICON)==0) return ;
  if (cw->icon_width>w || cw->icon_height>h) return ;

  if (w>cw->icon_width) x+= (w-cw->icon_width)/2;
  if (h>cw->icon_height) y+= (h-cw->icon_height)/2;

  blend_alpha_image( pager_image->data, pager_w, pager_h, x,y,
                     cw->icon_data, cw->icon_width, cw->icon_height, 0, 0,
                     cw->icon_width, cw->icon_height, bg);
}

static void pager_click(void *ignore, XButtonEvent *ev)
{
  int K;
  if (ev->button!=1) return ;
  K= ev->x / (page_w+1);
  if (K!=current_desktop)
    set_current_desktop(K);
}

static void pager_expose(void *ignore, XExposeEvent *ev)
{
  put_image(pager_window, ev->x, ev->y,
            pager_image, ev->x, ev->y,
            ev->width, ev->height);
}

static void pager_redraw()
{
  render_pager();
  XClearArea(x_display, pager_window, 0, 0, pager_w, pager_h, True);
}

static void root_property_change(void *ignore, XPropertyEvent *pe)
{
  //if (pe->state!=PropertyNewValue) return ;
  if (pe->atom==a_net_client_list_stacking)
  {
    pager_redraw();
    return ;
  }
  if (pe->atom==a_net_client_list)
  {
    taskbar_client_list_changed();
    return ;
  }
  if (pe->atom==a_net_current_desktop)
  {
    uint32_t cd;
    if (get_current_desktop(&cd)) return ;
    if (cd==current_desktop) return ;
    current_desktop= cd;
    pager_redraw();
    taskbar_current_desktop_changed();
    return ;
  }
 if (pe->atom==a_net_number_of_desktops)
  {
    uint32_t nd;
    if (get_number_of_desktops(&nd)) return ;
    if (nd==n_desktops) return ;
    n_desktops= nd;
    calculate_pager_dimensions(pager_h);
    destroy_image(pager_image);
    pager_image= create_image(pager_w, pager_h);
    XResizeWindow(x_display, pager_window, pager_w, pager_h);
    *pager_w_tgt= pager_w;
    pager_resized();
    pager_redraw();
    taskbar_number_of_desktops_changed();
    return ;
  }
  if (pe->atom==a_net_active_window)
  {
    taskbar_active_window_changed();
    return ;
  }
}

void client_size_change(client_window_t *cw, XConfigureEvent *ev)
{
  cw->flags &= ~CWF_POS; cw->tried &= ~CWF_POS;
  pager_redraw();
}

void client_property_change(client_window_t *cw, XPropertyEvent *pe)
{
   if (pe->atom==a_wm_name ||
       pe->atom==a_wm_icon_name ||
       pe->atom==a_net_wm_visible_icon_name ||
       pe->atom==a_net_wm_icon_name ||
       pe->atom==a_net_wm_name ||
       pe->atom==a_net_wm_visible_name)
   {
     cw->flags &= ~CWF_TITLE;
     cw->tried &= ~CWF_TITLE;
     if (cw->title) free(cw->title);
     cw->title= NULL;
     taskbar_window_title_changed(pe->window);
     return ;
   }
   if (pe->atom==a_net_wm_state)
   {
     cw->flags &= ~CWF_STATE;
     cw->tried &= ~CWF_STATE;
   }
   else if (pe->atom==a_net_wm_desktop)
   {
     cw->flags &= ~CWF_DESK;
     cw->tried &= ~CWF_DESK;
     taskbar_client_desktop_changed(pe->window);
   } else return ;
   pager_redraw();
}


