#include <stdint.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "x-basics.h"

Display *x_display;
Window   x_root_window;
int      x_screen_number;
uint32_t x_root_width, x_root_height;
uint32_t x_red_mask, x_red_shift,
                x_green_mask, x_green_shift,
                x_blue_mask, x_blue_shift;
int      x_depth;
Visual  *x_visual;

/***********************************************************************
  In order to create images etc, we need to find a suitable visual 
  with depth==24 or depth==32.
 ***********************************************************************/

static uint32_t trailing_zero_bits(uint32_t Y)
{
  uint32_t shif;
  for(shif=0;Y%2==0;shif++) Y= Y>>1;
  return shif;
}

static int get_x_visual()
{
  int depth, ndepths, i;
  int *depths;
  XVisualInfo xvinfo;

  depths= XListDepths(x_display, x_screen_number, &ndepths);
  if (!depths) 
  {
    fprintf(stderr,"can't get depth list for display\n");
    return 1;
  }

  depth=0;
  for(i=0;i<ndepths;i++)
    if (depths[i]==32 || depths[i]==24)
    {
      depth= depths[i];
      break;
    }
  XFree(depths);

  if (!depth)
  {
    fprintf(stderr, "display doesn't support 24 or 32 bit depth.\n");
    return 1;
  }

  x_depth= depth;
  x_visual= NULL;
  if (XMatchVisualInfo(x_display, x_screen_number, depth, TrueColor,&xvinfo))
     x_visual= xvinfo.visual;
  else if (XMatchVisualInfo(x_display, x_screen_number,
                                       depth, DirectColor,&xvinfo))
     x_visual= xvinfo.visual;
  else
  {
    fprintf(stderr,"can't find a 24 or 32 bit TrueColor/DirectColor visual\n");
    return 1;
  }
  x_red_mask= x_visual->red_mask;
  x_red_shift= trailing_zero_bits(x_red_mask);
  x_green_mask= x_visual->green_mask;
  x_green_shift= trailing_zero_bits(x_green_mask);
  x_blue_mask= x_visual->blue_mask;
  x_blue_shift= trailing_zero_bits(x_blue_mask);
  return 0;
}


int init_display(char *name)
{
  x_display= XOpenDisplay(name);

  if (!x_display) 
  {
    //FIXME: log this
    fprintf(stderr,"can't open display\n");
    return 1;
  }
  
  x_screen_number= DefaultScreen(x_display);
  x_root_window= DefaultRootWindow(x_display);

  if (get_x_visual())
  {
    XCloseDisplay(x_display);
    x_display= NULL;
    return 1;
  }

  x_root_width= DisplayWidth(x_display, x_screen_number);
  x_root_height= DisplayHeight(x_display, x_screen_number);

  return 0;
}

Window create_window_Ob
   (Window parent, int x,int y, int w, int h,
    unsigned int events, int border)
{
  Window win;
  XSetWindowAttributes attribs;

  attribs.background_pixmap= None;
  attribs.override_redirect= True;

  win= XCreateWindow(x_display, parent,
                     x, y, w, h,
                     border, x_depth, InputOutput, x_visual, 
                     CWBackPixmap | CWOverrideRedirect, &attribs);

  XSelectInput(x_display, win, events);
  XMapWindow(x_display, win);
  return win;
}
