#include <stdlib.h>
#include <X11/Xlib.h>
#include <stdint.h>
#include "x-basics.h"
#include "x-events.h"



typedef struct 
  { Window window; unsigned int event;
    void *data; void (*func)(void *, XEvent *); }
  event_handler_t;

static int n_event_handlers, s_event_handlers;
static event_handler_t *event_handlers;

static event_handler_t *find_event(Window window, unsigned int event);

void register_event_handler(Window window, unsigned int event,
                         void *data, void *func, void**old)
{
  event_handler_t *EV,*CP;
  EV= find_event(window, event);
  if (old) *old= NULL;
  if (EV)
  {
    if (old)
    {
      CP= malloc(sizeof(*CP));
      *CP= *EV;
      *old= CP;
    }
    goto done;
  }

  if (n_event_handlers==s_event_handlers)
  {
    if (s_event_handlers==0)
    {
      s_event_handlers= 20;
      event_handlers= malloc(sizeof(event_handlers[0])*s_event_handlers);
    }
    else
    {
      s_event_handlers+= s_event_handlers/2;
      event_handlers= realloc(event_handlers,
                              sizeof(event_handlers[0])*s_event_handlers);
    }
  }
  EV= event_handlers+n_event_handlers;
  n_event_handlers++;

done:
  EV->window= window;
  EV->event= event;
  EV->data= data;
  EV->func= func;
}

void unregister_all_event_handlers(Window window)
{
  int i;
  for(i=0;i<n_event_handlers;i++)
    if (event_handlers[i].window==window)
    {
      event_handlers[i]= event_handlers[n_event_handlers-1];
      n_event_handlers--; i--;
    }
}

void unregister_event_handler(Window window, unsigned int event, void *old)
{
  event_handler_t *EV,*NE;
  EV= old;
  NE= find_event(window, event);
  if (NE && EV)         *NE= *EV; 
  else if (NE && !EV)   *NE= event_handlers[--n_event_handlers];
  else if (!NE && EV)   register_event_handler(window, event,
                                               EV->data, EV->func,NULL);
  if (EV) free(EV);
}


static event_handler_t *find_event(Window window, unsigned int event)
{
  event_handler_t *EV;
  int i;
  for(i=0;i<n_event_handlers;i++)
  {
    EV= event_handlers+i;
    if (EV->window==window && EV->event==event)
       return EV;
  }
  return NULL;
}

#include <syslog.h>

void handle_registered_event(XEvent *event)
{
  event_handler_t *EV;

  if (event->xany.window==x_root_window)
  {  if (event->type==ButtonPress) syslog(LOG_USER|LOG_NOTICE,"root press\n");
    if (event->type==ButtonRelease) syslog(LOG_USER|LOG_NOTICE,"root release\n");
 }
                              

  EV= find_event(event->xany.window, event->type);
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void call_event_handler(void *H,XEvent *event)
{
  event_handler_t *EV;
  EV= H;
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void free_event_handler(void *H)
{
  if (H) free(H);
}


