#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>
#include <X11/Xutil.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "txtren.h"
#include "graymap.h"

#include "txent.h"
#include "utf8.h"


typedef struct { uint32_t chr; uint8_t  *d; uint32_t w; int x; } glyph_t;

struct xtext
 { Window window; ximage_t *image; int x,y,w,h;
   glyph_t *glyphs; int size, len;
   int has_focus; uint32_t left; int cursor; };

static u_font_t *te_font;

static uint32_t TE_fg        = 0;
static uint32_t TE_bg        = 0xffffff;
static uint32_t TE_cfg       = 0xffffff;
static uint32_t TE_cbg       = 0x000088;
static int      TE_font_size = 12;
static char    *TE_font_name = ":default";


static void mk_glyph(glyph_t *G, uint32_t chr);
static void render_te(text_entry_t *te);
static void make_cursor_visible(text_entry_t *te, int redraw);
static void convert_to_wide_str(char *str, uint32_t **Rwstr, uint32_t *Rwlen);
static int draw_glyph(text_entry_t  *te, int gn, int expose);
static void move_cursor(text_entry_t *te, int newcurs);
static void insert_char(text_entry_t *te, uint32_t chr);
static void delete_cursor_char(text_entry_t *te);

text_entry_t *text_entry_init
    (Window window, ximage_t *image, int x,int y, int w, int h,
        char *initstr,int focus)
{
  text_entry_t *te;
  uint32_t i;
  uint32_t *wide_str, wide_str_len;
  int size;

  if (!te_font) te_font= font_new(TE_font_name, TE_font_size);

  te= malloc(sizeof(*te));
  te->window= window; te->image= image;
  te->x= x; te->y= y; te->w= w; te->h= h;
  te->has_focus= focus;

  if (!initstr) initstr= "";
  convert_to_wide_str(initstr, &wide_str, &wide_str_len);

  size= wide_str_len + wide_str_len/2 + 2;
  if (size<50) size= 50;

  te->size= size;
  te->len= wide_str_len+1;
  te->glyphs= malloc(sizeof(te->glyphs[0])*te->size);
  
  for(i=0;i<te->len-1;i++) mk_glyph(te->glyphs+i, wide_str[i]);
  mk_glyph(te->glyphs+i, ' ');

  x= 0;
  for(i=0;i<te->len;i++)
  {
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  
  free(wide_str);
  
  te->cursor= te->len-1;
  te->left= 0;
  make_cursor_visible(te, 1);
  return te;
}

static void convert_to_wide_str(char *str, uint32_t **Rwstr, uint32_t *Rwlen)
{
  uint32_t *wstr; 
  size_t wlen;
  uint32_t slen, i,j;

  slen= strlen(str);
  wstr= malloc((slen+1)*sizeof(uint32_t));

  if (utf8_decode_string((uint8_t*) str, slen, wstr, &wlen))
  {
    j= 0;
    for(i=0;i<wlen;i++)
    {
     if (wstr[i]&(1<<31)) continue;
     wstr[j]= wstr[i];
     j++;
    }
    wlen= j;
  }
  *Rwstr= wstr;
  *Rwlen= wlen;
}

static void make_cursor_visible(text_entry_t *te, int redraw)
{
  int cx, cw;
  cx= te->glyphs[te->cursor].x;
  cw= te->glyphs[te->cursor].w;

  if (cx<te->left) { te->left= cx; goto render; }
  if (cx+cw-te->left>te->w) { te->left= cx+cw-te->w; goto render; }
  if (!redraw) return ;
render:
  render_te(te);
}

static void render_te(text_entry_t *te)
{
  int i;
  fill_rect(te->image->data, te->image->width, te->image->height,
            te->x,te->y, te->w, te->h, TE_bg);
  for(i=0;i<te->len;i++) if (draw_glyph(te, i, 0)) break;
  XClearArea(x_display, te->window, te->x, te->y, te->w, te->h, True);
}

static void mk_glyph(glyph_t *G, uint32_t chr)
{
  text_render_wide(te_font, &chr, 1, &G->w, &G->d);
  G->chr= chr;
}


void text_entry_focus(text_entry_t *te,int focus)
{
  if (te->has_focus==focus) return ;
  te->has_focus= focus;
  draw_glyph(te, te->cursor, 1);
}

void text_entry_key(text_entry_t *te, KeySym sym, uint32_t *chr, size_t nchr)
{
  switch(sym)
  {
   case XK_Left:
     if (te->cursor!=0) move_cursor(te, te->cursor-1);
     break;
   case XK_Right:
     if (te->cursor!=te->len-1) move_cursor(te,te->cursor+1);
     break;
   case XK_Home: move_cursor(te, 0); break;
   case XK_End:  move_cursor(te, te->len-1); break;
   case XK_Delete:
     if (te->cursor!=te->len-1) delete_cursor_char(te);
     break;
   case XK_BackSpace:
     if (te->cursor!=0)
     {
       te->cursor--;
       delete_cursor_char(te);
     }
     break;
   default:
     if (nchr && chr && chr[0])
        insert_char(te, chr[0]);
  }
}

static int draw_glyph(text_entry_t  *te, int gn, int expose)
{
  int ws, we, sx, sy, cw, ch, dx, gs,ge;
  uint32_t fg, bg;
  glyph_t *G;

  ws= te->left; 
  we= te->left+te->w;
  G= te->glyphs+gn;

  sy= 0; ch= te_font->height;
  if (te_font->height>te->h)
  {
    sy= (te_font->height - te->h)/2;
    ch= te->h;
  }

  gs= G->x;
  ge= gs + G->w;

  if (ge<=ws) return 0;
  if (gs>=we) return 1;
    
  sx= 0; cw= te->glyphs[gn].w;
  if (gs<ws) { sx= ws-gs; cw-= ws-gs; }
  if (ge>we) { cw-= ge-we; }

  dx= sx + gs - ws;
  if (te->has_focus && gn==te->cursor) { fg= TE_cfg; bg= TE_cbg; }
                                  else { fg= TE_fg;  bg= TE_bg; }
  graymap_draw( te->image->data, te->image->width, te->image->height,
                                 dx + te->x, te->y,
                G->d, G->w, te_font->height,
                                 sx, sy, fg, bg, cw, ch); 
  if (expose)
    XClearArea(x_display, te->window, dx+te->x, te->y, cw, ch, True);
  return 0;
}

static void move_cursor(text_entry_t *te, int newcurs)
{
  int ws, we, gs,ge;
  int C;

  if (newcurs<0 || newcurs>=te->len || newcurs==te->cursor) return ;  

  gs= te->glyphs[newcurs].x;
  ge= gs + te->glyphs[newcurs].w;

  ws= te->left; 
  we= te->left+te->w;

  if (gs<ws|| ge>we)
  {
    te->cursor= newcurs;
    make_cursor_visible(te, 1);
    return ;
  }

  C= te->cursor;
  te->cursor= newcurs;
  draw_glyph(te, C,1);
  draw_glyph(te, te->cursor,1);
}

static void insert_char(text_entry_t *te, uint32_t chr)
{
  int i;
  if (te->len==te->size)
  {
    te->size+= te->size/2;
    te->glyphs= realloc(te->glyphs, sizeof(te->glyphs[0])*te->size);
  }
  for(i=te->len;i>te->cursor;i--)
    te->glyphs[i]= te->glyphs[i-1];
  mk_glyph(te->glyphs+ te->cursor, chr);
  
  te->cursor++;
  te->len++;
  for(i=te->cursor;i<te->len;i++)
    te->glyphs[i].x= te->glyphs[i-1].x + te->glyphs[i-1].w;

  make_cursor_visible(te, 1);
}

static void delete_cursor_char(text_entry_t *te)
{
  int i,x;
  free(te->glyphs[te->cursor].d);
  x= te->glyphs[te->cursor].x;
  for(i=te->cursor;i<te->len-1;i++)
  {
    te->glyphs[i]= te->glyphs[i+1];
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  te->len--;
  te->glyphs[i].x= x;
  make_cursor_visible(te, 1);
}

char *text_entry_value(text_entry_t *te)
{
  uint32_t *wide, widelen;
  uint8_t *str;
  int i;
  size_t slen;

  widelen= te->len;
  wide= malloc(sizeof(uint32_t)*widelen);
  for(i=0;i<te->len;i++)
    wide[i]= te->glyphs[i].chr;

  str= malloc(widelen+1);
  utf8_encode_string(wide, widelen, str, &slen);
  if (slen) str[slen-1]= 0; // eat up the space we added
  free(wide);
  return (char*) str;
}

void text_entry_free(text_entry_t *te)
{
  int i;
  for(i=0;i<te->len;i++)
   free(te->glyphs[i].d);
  free(te->glyphs);
  free(te);
}

void text_entry_set(text_entry_t *te, char *str)
{
  uint32_t i;
  uint32_t *wide_str, wide_str_len;
  int x;

  convert_to_wide_str(str, &wide_str, &wide_str_len);

  if (wide_str_len>te->size)
  {
    te->size= wide_str_len + wide_str_len/2 + 2;
    if (te->size<50) te->size= 50;
    te->glyphs= realloc(te->glyphs,sizeof(te->glyphs[0])*te->size);
  }

  for(i=0;i<te->len;i++) free(te->glyphs[i].d);

  te->len= wide_str_len+1;
  
  for(i=0;i<te->len-1;i++) mk_glyph(te->glyphs+i, wide_str[i]);
  mk_glyph(te->glyphs+i, ' ');

  x= 0;
  for(i=0;i<te->len;i++)
  {
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  
  free(wide_str);
  
  te->cursor= te->len-1;
  te->left= 0;
  make_cursor_visible(te, 1);
}

